"""
Sklearn similarity backend.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np
import sklearn.metrics as skm
import sklearn.neighbors as skn
import sklearn.preprocessing as skp

import eta.core.utils as etau

from fiftyone.brain.similarity import (
    DuplicatesMixin,
    SimilarityConfig,
    Similarity,
    SimilarityIndex,
)
import fiftyone.brain.internal.core.utils as fbu


logger = logging.getLogger(__name__)

_AGGREGATIONS = {
    "mean": np.mean,
    "post-mean": np.mean,
    "post-min": np.min,
    "post-max": np.max,
}

_MAX_PRECOMPUTE_DISTS = 15000  # ~1.7GB to store distance matrix in-memory
_COSINE_HACK_ATTR = "_cosine_hack"


class SklearnSimilarityConfig(SimilarityConfig):
    """Configuration for the sklearn similarity backend.

    Args:
        embeddings_field (None): the sample field containing the embeddings,
            if one was provided
        model (None): the :class:`fiftyone.core.models.Model` or name of the
            zoo model that was used to compute embeddings, if known
        patches_field (None): the sample field defining the patches being
            analyzed, if any
        supports_prompts (None): whether this run supports prompt queries
        metric ("cosine"): the embedding distance metric to use. See
            ``sklearn.metrics.pairwise_distance`` for supported values
    """

    def __init__(
        self,
        embeddings_field=None,
        model=None,
        patches_field=None,
        supports_prompts=None,
        metric="cosine",
        **kwargs,
    ):
        super().__init__(
            embeddings_field=embeddings_field,
            model=model,
            patches_field=patches_field,
            supports_prompts=supports_prompts,
            **kwargs,
        )
        self.metric = metric

    @property
    def method(self):
        return "sklearn"

    @property
    def max_k(self):
        return None

    @property
    def supports_least_similarity(self):
        return True

    @property
    def supported_aggregations(self):
        return tuple(_AGGREGATIONS.keys())


class SklearnSimilarity(Similarity):
    """Sklearn similarity factory.

    Args:
        config: an :class:`SklearnSimilarityConfig`
    """

    def initialize(self, samples, brain_key):
        return SklearnSimilarityIndex(
            samples, self.config, brain_key, backend=self
        )


class SklearnSimilarityIndex(SimilarityIndex, DuplicatesMixin):
    """Class for interacting with sklearn similarity indexes.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`SklearnSimilarityConfig` used
        brain_key: the brain key
        embeddings (None): a ``num_embeddings x num_dims`` array of embeddings
        sample_ids (None): a ``num_embeddings`` array of sample IDs
        label_ids (None): a ``num_embeddings`` array of label IDs, if
            applicable
        backend (None): a :class:`SklearnSimilarity` instance
    """

    def __init__(
        self,
        samples,
        config,
        brain_key,
        embeddings=None,
        sample_ids=None,
        label_ids=None,
        backend=None,
    ):
        embeddings, sample_ids, label_ids = self._parse_data(
            samples,
            config,
            embeddings=embeddings,
            sample_ids=sample_ids,
            label_ids=label_ids,
        )

        self._embeddings = embeddings
        self._sample_ids = sample_ids
        self._label_ids = label_ids

        self._ids_to_inds = None
        self._curr_ids_to_inds = None
        self._neighbors_helper = None

        SimilarityIndex.__init__(
            self, samples, config, brain_key, backend=backend
        )
        DuplicatesMixin.__init__(self)

    @property
    def embeddings(self):
        return self._embeddings

    @property
    def sample_ids(self):
        return self._sample_ids

    @property
    def label_ids(self):
        return self._label_ids

    @property
    def total_index_size(self):
        return len(self._sample_ids)

    def add_to_index(
        self,
        embeddings,
        sample_ids,
        label_ids=None,
        overwrite=True,
        allow_existing=True,
        warn_existing=False,
    ):
        _sample_ids, _label_ids, ii, jj = fbu.add_ids(
            sample_ids,
            label_ids,
            self._sample_ids,
            self._label_ids,
            patches_field=self.config.patches_field,
            overwrite=overwrite,
            allow_existing=allow_existing,
            warn_existing=warn_existing,
        )

        if ii.size == 0:
            return

        _embeddings = embeddings[ii, :]

        if self.config.embeddings_field is not None:
            fbu.add_embeddings(
                self._samples,
                _embeddings,
                _sample_ids,
                _label_ids,
                self.config.embeddings_field,
                patches_field=self.config.patches_field,
            )

        _e = self._embeddings

        n = _e.shape[0]
        if n == 0:
            _e = np.empty((0, embeddings.shape[1]), dtype=embeddings.dtype)

        d = _e.shape[1]
        m = jj[-1] - n + 1

        if m > 0:
            if _e.size > 0:
                _e = np.concatenate((_e, np.empty((m, d), dtype=_e.dtype)))
            else:
                _e = np.empty_like(_embeddings)

        _e[jj, :] = _embeddings

        self._embeddings = _e
        self._sample_ids = _sample_ids
        self._label_ids = _label_ids

    def remove_from_index(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        _sample_ids, _label_ids, rm_inds = fbu.remove_ids(
            sample_ids,
            label_ids,
            self._sample_ids,
            self._label_ids,
            patches_field=self.config.patches_field,
            allow_missing=allow_missing,
            warn_missing=warn_missing,
        )

        if rm_inds.size == 0:
            return

        if self.config.embeddings_field is not None:
            fbu.remove_embeddings(
                self._samples,
                self.config.embeddings_field,
                sample_ids=_sample_ids,
                label_ids=_label_ids,
                patches_field=self.config.patches_field,
            )

        _embeddings = np.delete(self._embeddings, rm_inds)

        self._embeddings = _embeddings
        self._sample_ids = _sample_ids
        self._label_ids = _label_ids

    def use_view(self, *args, **kwargs):
        self._curr_ids_to_inds = None
        return super().use_view(*args, **kwargs)

    def get_embeddings(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        if label_ids is not None:
            if self.config.patches_field is None:
                raise ValueError("This index does not support label IDs")

            if sample_ids is not None:
                logger.warning(
                    "Ignoring sample IDs when label IDs are provided"
                )

            inds = _get_inds(
                label_ids,
                self.label_ids,
                "label",
                allow_missing,
                warn_missing,
            )

            embeddings = self._embeddings[inds, :]
            sample_ids = self.sample_ids[inds]
            label_ids = np.asarray(label_ids)
        elif sample_ids is not None:
            if etau.is_str(sample_ids):
                sample_ids = [sample_ids]

            if self.config.patches_field is not None:
                sample_ids = set(sample_ids)
                bools = [_id in sample_ids for _id in self.sample_ids]
                inds = np.nonzero(bools)[0]
            else:
                inds = _get_inds(
                    sample_ids,
                    self.sample_ids,
                    "sample",
                    allow_missing,
                    warn_missing,
                )

            embeddings = self._embeddings[inds, :]
            sample_ids = self.sample_ids[inds]
            if self.config.patches_field is not None:
                label_ids = self.label_ids[inds]
            else:
                label_ids = None
        else:
            embeddings = self._embeddings.copy()
            sample_ids = self.sample_ids.copy()
            if self.config.patches_field is not None:
                label_ids = self.label_ids.copy()
            else:
                label_ids = None

        return embeddings, sample_ids, label_ids

    def reload(self):
        if self.config.embeddings_field is not None:
            embeddings, sample_ids, label_ids = self._parse_data(
                self._samples, self.config
            )

            self._embeddings = embeddings
            self._sample_ids = sample_ids
            self._label_ids = label_ids

            self._ids_to_inds = None
            self._curr_ids_to_inds = None
            self._neighbors_helper = None

        super().reload()

    def cleanup(self):
        pass

    def attributes(self):
        attrs = super().attributes()

        if self.config.embeddings_field is None:
            attrs.extend(["embeddings", "sample_ids", "label_ids"])

        return attrs

    def _kneighbors(
        self,
        query=None,
        k=None,
        reverse=False,
        aggregation=None,
        return_dists=False,
    ):
        if aggregation is not None:
            return self._kneighbors_aggregate(
                query, k, reverse, aggregation, return_dists
            )

        (
            query,
            query_inds,
            full_index,
            single_query,
        ) = self._parse_neighbors_query(query)

        can_use_dists = full_index or query_inds is not None
        neighbors, dists = self._get_neighbors(can_use_dists=can_use_dists)

        if dists is not None:
            # Use pre-computed distances
            if query_inds is not None:
                _dists = dists[query_inds, :]
                _rows = range(len(query_inds))
            else:
                _dists = dists
                _rows = range(dists.shape[0])

            inds = np.argmin(_dists, axis=1)
            if return_dists:
                dists = _dists[_rows, inds]
            else:
                dists = None
        else:
            if return_dists:
                dists, inds = neighbors.kneighbors(
                    X=query, n_neighbors=k, return_distance=True
                )
            else:
                inds = neighbors.kneighbors(
                    X=query, n_neighbors=k, return_distance=False
                )
                dists = None

        return self._format_output(
            inds, dists, full_index, single_query, return_dists
        )

    def _radius_neighbors(self, query=None, thresh=None, return_dists=False):
        (
            query,
            query_inds,
            full_index,
            single_query,
        ) = self._parse_neighbors_query(query)

        can_use_dists = full_index or query_inds is not None
        neighbors, dists = self._get_neighbors(can_use_dists=can_use_dists)

        # When not using brute force, we approximate cosine distance by
        # computing Euclidean distance on unit-norm embeddings.
        # ED = sqrt(2 * CD), so we need to scale the threshold appropriately
        if getattr(neighbors, _COSINE_HACK_ATTR, False):
            thresh = np.sqrt(2.0 * thresh)

        if dists is not None:
            # Use pre-computed distances
            if query_inds is not None:
                _dists = dists[query_inds, :]
            else:
                _dists = dists

            inds = [np.nonzero(d <= thresh)[0] for d in _dists]
            if return_dists:
                dists = [d[i] for i, d in zip(inds, _dists)]
            else:
                dists = None
        else:
            if return_dists:
                dists, inds = neighbors.radius_neighbors(
                    X=query, radius=thresh, return_distance=True
                )
            else:
                dists = None
                inds = neighbors.radius_neighbors(
                    X=query, radius=thresh, return_distance=False
                )

        return self._format_output(
            inds, dists, full_index, single_query, return_dists
        )

    def _kneighbors_aggregate(
        self, query, k, reverse, aggregation, return_dists
    ):
        if query is None:
            raise ValueError("Full index queries do not support aggregation")

        if aggregation not in _AGGREGATIONS:
            raise ValueError(
                "Unsupported aggregation method '%s'. Supported values are %s"
                % (aggregation, tuple(_AGGREGATIONS.keys()))
            )

        query, query_inds, _, _ = self._parse_neighbors_query(query)

        # Pre-aggregation
        if aggregation == "mean":
            if query.shape[0] > 1:
                query = query.mean(axis=0, keepdims=True)
                query_inds = None

            aggregation = None

        can_use_dists = query_inds is not None
        _, dists = self._get_neighbors(
            can_use_neighbors=False, can_use_dists=can_use_dists
        )

        if dists is not None:
            # Use pre-computed distances
            dists = dists[query_inds, :]
        else:
            keep_inds = self._current_inds
            index_embeddings = self._embeddings
            if keep_inds is not None:
                index_embeddings = index_embeddings[keep_inds]

            dists = skm.pairwise_distances(
                query, index_embeddings, metric=self.config.metric
            )

        # Post-aggregation
        if aggregation is not None:
            agg_fcn = _AGGREGATIONS[aggregation]
            dists = agg_fcn(dists, axis=0)
        else:
            dists = dists[0, :]

        inds = np.argsort(dists)
        if reverse:
            inds = np.flip(inds)

        if k is not None:
            inds = inds[:k]

        if self.config.patches_field is not None:
            ids = self.current_label_ids
        else:
            ids = self.current_sample_ids

        if return_dists:
            return ids[inds], dists[inds]

        return ids[inds]

    def _parse_neighbors_query(self, query):
        # Full index
        if query is None:
            return None, None, True, False

        if etau.is_str(query):
            query_ids = [query]
            single_query = True
        else:
            query = np.asarray(query)

            # Query vector(s)
            if np.issubdtype(query.dtype, np.number):
                single_query = query.ndim == 1
                if single_query:
                    query = query[np.newaxis, :]

                return query, None, False, single_query

            query_ids = list(query)
            single_query = False

        # Retrieve indices into active `dists` matrix, if possible
        ids_to_inds = self._get_ids_to_inds(full=False)
        query_inds = []
        for _id in query_ids:
            _ind = ids_to_inds.get(_id, None)
            if _ind is not None:
                query_inds.append(_ind)
            else:
                # At least one query ID is not in the active index
                query_inds = None
                break

        # Retrieve embeddings
        ids_to_inds = self._get_ids_to_inds(full=True)
        inds = []
        bad_ids = []
        for _id in query_ids:
            _ind = ids_to_inds.get(_id, None)
            if _ind is not None:
                inds.append(_ind)
            else:
                bad_ids.append(_id)

        inds = np.array(inds)

        if bad_ids:
            raise ValueError(
                "Query IDs %s do not exist in this index" % bad_ids
            )

        query = self._embeddings[inds, :]

        if query_inds is not None:
            query_inds = np.array(query_inds)

        return query, query_inds, False, single_query

    def _get_ids_to_inds(self, full=False):
        if full:
            if self._ids_to_inds is None:
                if self.config.patches_field is not None:
                    ids = self.label_ids
                else:
                    ids = self.sample_ids

                self._ids_to_inds = {_id: i for i, _id in enumerate(ids)}

            return self._ids_to_inds

        if self._curr_ids_to_inds is None:
            if self.config.patches_field is not None:
                ids = self.current_label_ids
            else:
                ids = self.current_sample_ids

            self._curr_ids_to_inds = {_id: i for i, _id in enumerate(ids)}

        return self._curr_ids_to_inds

    def _get_neighbors(self, can_use_neighbors=True, can_use_dists=True):
        if self._neighbors_helper is None:
            self._neighbors_helper = NeighborsHelper(
                self._embeddings, self.config.metric
            )

        return self._neighbors_helper.get_neighbors(
            keep_inds=self._current_inds,
            can_use_neighbors=can_use_neighbors,
            can_use_dists=can_use_dists,
        )

    def _format_output(
        self, inds, dists, full_index, single_query, return_dists
    ):
        if full_index:
            return (inds, dists) if return_dists else inds

        if self.config.patches_field is not None:
            index_ids = self.current_label_ids
        else:
            index_ids = self.current_sample_ids

        ids = [[index_ids[i] for i in _inds] for _inds in inds]
        if return_dists:
            dists = [list(d) for d in dists]

        if single_query:
            ids = ids[0]
            if return_dists:
                dists = dists[0]

        return (ids, dists) if return_dists else ids

    @staticmethod
    def _parse_data(
        samples,
        config,
        embeddings=None,
        sample_ids=None,
        label_ids=None,
    ):
        if embeddings is None:
            embeddings, sample_ids, label_ids = fbu.get_embeddings(
                samples._dataset,
                patches_field=config.patches_field,
                embeddings_field=config.embeddings_field,
            )
        elif sample_ids is None:
            sample_ids, label_ids = fbu.get_ids(
                samples,
                patches_field=config.patches_field,
                data=embeddings,
                data_type="embeddings",
            )

        return embeddings, sample_ids, label_ids

    @classmethod
    def _from_dict(cls, d, samples, config, brain_key):
        embeddings = d.get("embeddings", None)
        if embeddings is not None:
            embeddings = np.array(embeddings)

        sample_ids = d.get("sample_ids", None)
        if sample_ids is not None:
            sample_ids = np.array(sample_ids)

        label_ids = d.get("label_ids", None)
        if label_ids is not None:
            label_ids = np.array(label_ids)

        return cls(
            samples,
            config,
            brain_key,
            embeddings=embeddings,
            sample_ids=sample_ids,
            label_ids=label_ids,
        )


class NeighborsHelper(object):

    _UNAVAILABLE = "UNAVAILABLE"

    def __init__(self, embeddings, metric):
        self.embeddings = embeddings
        self.metric = metric

        self._initialized = False
        self._full_dists = None

        self._curr_keep_inds = None
        self._curr_neighbors = None
        self._curr_dists = None

    def get_neighbors(
        self,
        keep_inds=None,
        can_use_neighbors=True,
        can_use_dists=True,
    ):
        iokay = self._same_keep_inds(keep_inds)
        nokay = not can_use_neighbors or self._curr_neighbors is not None
        dokay = not can_use_dists or self._curr_dists is not None

        if iokay and nokay and dokay:
            neighbors = self._curr_neighbors
            dists = self._curr_dists
        else:
            neighbors, dists = self._build(
                keep_inds=keep_inds,
                can_use_neighbors=can_use_neighbors,
                can_use_dists=can_use_dists,
            )

            if not iokay:
                self._curr_keep_inds = keep_inds

            if self._curr_neighbors is None or not iokay:
                self._curr_neighbors = neighbors

            if self._curr_dists is None or not iokay:
                self._curr_dists = dists

        if not can_use_neighbors or neighbors is self._UNAVAILABLE:
            neighbors = None

        if not can_use_dists or dists is self._UNAVAILABLE:
            dists = None

        return neighbors, dists

    def _same_keep_inds(self, keep_inds):
        # This handles either argument being None
        return np.array_equal(keep_inds, self._curr_keep_inds)

    def _build(
        self, keep_inds=None, can_use_neighbors=True, can_use_dists=True
    ):
        if can_use_dists:
            if (
                self._full_dists is None
                and len(self.embeddings) <= _MAX_PRECOMPUTE_DISTS
            ):
                self._full_dists = self._build_dists(self.embeddings)

            if self._full_dists is not None:
                if keep_inds is not None:
                    dists = self._full_dists[keep_inds, :][:, keep_inds]
                else:
                    dists = self._full_dists
            elif (
                keep_inds is not None
                and len(keep_inds) <= _MAX_PRECOMPUTE_DISTS
            ):
                dists = self._build_dists(self.embeddings[keep_inds])
            else:
                dists = self._UNAVAILABLE
        else:
            dists = None

        if can_use_neighbors:
            if not isinstance(dists, np.ndarray):
                embeddings = self.embeddings
                if keep_inds is not None:
                    embeddings = embeddings[keep_inds]

                neighbors = self._build_neighbors(embeddings)
            else:
                neighbors = self._UNAVAILABLE
        else:
            neighbors = None

        return neighbors, dists

    def _build_dists(self, embeddings):
        logger.info("Generating index for %d embeddings...", len(embeddings))

        # Center embeddings
        embeddings = np.asarray(embeddings)
        embeddings -= embeddings.mean(axis=0, keepdims=True)

        dists = skm.pairwise_distances(embeddings, metric=self.metric)

        logger.info("Index complete")

        return dists

    def _build_neighbors(self, embeddings):
        logger.info(
            "Generating neighbors graph for %d embeddings...",
            len(embeddings),
        )

        # Center embeddings
        embeddings = np.asarray(embeddings)
        embeddings -= embeddings.mean(axis=0, keepdims=True)

        metric = self.metric

        if metric == "cosine":
            # Nearest neighbors does not directly support cosine distance, so
            # we approximate via euclidean distance on unit-norm embeddings
            cosine_hack = True
            embeddings = skp.normalize(embeddings, axis=1)
            metric = "euclidean"
        else:
            cosine_hack = False

        neighbors = skn.NearestNeighbors(metric=metric)
        neighbors.fit(embeddings)

        setattr(neighbors, _COSINE_HACK_ATTR, cosine_hack)

        logger.info("Index complete")

        return neighbors


def _get_inds(ids, index_ids, ftype, allow_missing, warn_missing):
    if etau.is_str(ids):
        ids = [ids]

    ids_map = {_id: i for i, _id in enumerate(index_ids)}

    inds = []
    bad_ids = []

    for _id in ids:
        idx = ids_map.get(_id, None)
        if idx is not None:
            inds.append(idx)
        else:
            bad_ids.append(_id)

    num_missing = len(bad_ids)

    if num_missing > 0:
        if not allow_missing:
            raise ValueError(
                "Found %d %s IDs (eg '%s') that are not present in the index"
                % (num_missing, ftype, bad_ids[0])
            )

        if warn_missing:
            logger.warning(
                "Ignoring %d %s IDs that are not present in the index",
                num_missing,
                ftype,
            )

    return np.array(inds)



from pkgutil import extend_path

#
# This statement allows multiple `fiftyone.XXX` packages to be installed in the
# same environment and used simultaneously.
#
# https://docs.python.org/3/library/pkgutil.html#pkgutil.extend_path
#
__path__ = extend_path(__path__, __name__)

from fiftyone.__public__ import *



"""
Preprocessing functions

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import copy
from functools import singledispatch

import numpy as np
import torch
from torch import nn


## Data Preprocessing and Handling
def preprocess(dataset, transforms):
    dataset = copy.copy(dataset)  # shallow copy
    for transform in transforms:
        dataset["data"] = transform(dataset["data"])
    return dataset


@singledispatch
def normalise(x, mean, std):
    return (x - mean) / std


@normalise.register(np.ndarray)
def _(x, mean, std):
    # faster inplace for numpy arrays
    x = np.array(x, np.float32)
    x -= mean
    x *= 1.0 / std
    return x


unnormalise = lambda x, mean, std: x * std + mean


@singledispatch
def pad(x, border):
    raise NotImplementedError


@pad.register(np.ndarray)
def _(x, border):
    return np.pad(
        x, [(0, 0), (border, border), (border, border), (0, 0)], mode="reflect"
    )


@pad.register(torch.Tensor)
def _(x, border):
    return nn.ReflectionPad2d(border)(x)


@singledispatch
def transpose(x, source, target):
    raise NotImplementedError


@transpose.register(np.ndarray)
def _(x, source, target):
    return x.transpose([source.index(d) for d in target])


@transpose.register(torch.Tensor)
def _(x, source, target):
    return x.permute([source.index(d) for d in target])


class Transform:
    def __init__(self, dataset, transforms):
        self.dataset, self.transforms = dataset, transforms
        self.choices = None

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        data, labels = self.dataset[index]
        data = data.copy()
        for choices, f in zip(self.choices, self.transforms):
            data = f(data, **choices[index])
        return data, labels

    def set_random_choices(self):
        self.choices = []
        x_shape = self.dataset[0][0].shape
        N = len(self)
        for t in self.transforms:
            self.choices.append(np.random.choice(t.options(x_shape), N))
            x_shape = (
                t.output_shape(x_shape)
                if hasattr(t, "output_shape")
                else x_shape
            )



"""
Utilities.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import itertools
import logging
import random
import string

import numpy as np

import eta.core.utils as etau

import fiftyone.brain as fob
import fiftyone.core.fields as fof
import fiftyone.core.labels as fol
import fiftyone.core.patches as fop
import fiftyone.zoo as foz
from fiftyone import ViewField as F


logger = logging.getLogger(__name__)


def parse_data(
    samples,
    patches_field=None,
    data=None,
    data_type="embeddings",
    allow_missing=True,
    warn_missing=True,
):
    if isinstance(data, fob.SimilarityIndex):
        return get_embeddings_from_index(
            samples,
            data,
            patches_field=None,
            allow_missing=True,
            warn_missing=True,
        )

    if patches_field is None:
        if isinstance(data, dict):
            sample_ids, data = zip(*data.items())
            return np.array(data), np.array(sample_ids), None

        sample_ids, _ = get_ids(samples, data=data, data_type=data_type)
        return data, sample_ids, None

    if isinstance(data, dict):
        value = next(iter(data.values()), None)
        if isinstance(value, np.ndarray) and value.ndim == 1:
            label_ids, data = zip(*data.items())
            return _parse_label_data(
                samples,
                patches_field,
                label_ids,
                data,
                data_type,
                allow_missing,
                warn_missing,
            )

    sample_ids, label_ids = get_ids(
        samples,
        patches_field=patches_field,
        data=data,
        data_type=data_type,
    )

    return data, sample_ids, label_ids


def _parse_label_data(
    samples,
    patches_field,
    label_ids,
    data,
    data_type,
    allow_missing,
    warn_missing,
):
    if samples._is_patches:
        sample_id_path = "sample_id"
    else:
        sample_id_path = "id"

    label_type, label_id_path = samples._get_label_field_path(
        patches_field, "id"
    )
    is_list_field = issubclass(label_type, fol._LABEL_LIST_FIELDS)

    ref_sample_ids, ref_label_ids = samples._dataset.values(
        [sample_id_path, label_id_path]
    )

    if is_list_field:
        ids_map = {}
        for _sample_id, _lids in zip(ref_sample_ids, ref_label_ids):
            if _lids:
                for _label_id in _lids:
                    ids_map[_label_id] = _sample_id
    else:
        ids_map = dict(zip(ref_label_ids, ref_sample_ids))

    _data = []
    _sample_ids = []
    _label_ids = []
    _missing_ids = []
    for _lid, _d in zip(label_ids, data):
        _sid = ids_map.get(_lid, None)
        if _sid is not None:
            _data.append(_d)
            _sample_ids.append(_sid)
            _label_ids.append(_lid)
        else:
            _missing_ids.append(_lid)

    num_missing = len(_missing_ids)
    if num_missing > 0:
        if not allow_missing:
            raise ValueError(
                "Unable to retrieve sample IDs for %d label IDs (eg %s)"
                % (num_missing, _missing_ids[0])
            )

        if warn_missing:
            logger.warning(
                "Ignoring %s for %d label IDs (eg %s) for which sample IDs "
                "could not be retrieved",
                data_type,
                num_missing,
                _missing_ids[0],
            )

    return np.array(_data), np.array(_sample_ids), np.array(_label_ids)


def get_embeddings_from_index(
    samples,
    similarity_index,
    patches_field=None,
    allow_missing=True,
    warn_missing=True,
):
    if patches_field is None:
        if samples._is_patches:
            sample_id_path = "sample_id"
        else:
            sample_id_path = "id"

        sample_ids = samples.values(sample_id_path)
        label_ids = None
    else:
        if samples._is_patches:
            label_id_path = "id"
        else:
            _, label_id_path = samples._get_label_field_path(
                patches_field, "id"
            )

        sample_ids = None
        label_ids = samples.values(label_id_path, unwind=True)

    return similarity_index.get_embeddings(
        sample_ids=sample_ids,
        label_ids=label_ids,
        allow_missing=allow_missing,
        warn_missing=warn_missing,
    )


def get_ids(
    samples,
    patches_field=None,
    data=None,
    data_type="embeddings",
    handle_missing="skip",
    ref_sample_ids=None,
):
    if patches_field is None:
        if ref_sample_ids is not None:
            sample_ids = ref_sample_ids
        else:
            sample_ids = samples.values("id")

        if data is not None and len(sample_ids) != len(data):
            raise ValueError(
                "The number of %s (%d) in these results no longer matches the "
                "number of samples (%d) in the collection. You must "
                "regenerate the results"
                % (data_type, len(data), len(sample_ids))
            )

        return np.array(sample_ids), None

    sample_ids, label_ids = _get_patch_ids(
        samples,
        patches_field,
        handle_missing=handle_missing,
        ref_sample_ids=ref_sample_ids,
    )

    if data is not None and len(sample_ids) != len(data):
        raise ValueError(
            "The number of %s (%d) in these results no longer matches the "
            "number of labels (%d) in the '%s' field of the collection. You "
            "must regenerate the results"
            % (data_type, len(data), len(sample_ids), patches_field)
        )

    return np.array(sample_ids), np.array(label_ids)


def filter_ids(
    samples,
    index_sample_ids,
    index_label_ids,
    patches_field=None,
    allow_missing=True,
    warn_missing=False,
):
    _validate_args(samples, None, patches_field)

    if patches_field is None:
        if samples._is_patches:
            sample_ids = np.array(samples.values("sample_id"))
        else:
            sample_ids = np.array(samples.values("id"))

        if index_sample_ids is None:
            return sample_ids, None, None, None

        keep_inds, good_inds, bad_ids = _parse_ids(
            sample_ids,
            index_sample_ids,
            "samples",
            allow_missing,
            warn_missing,
        )

        if bad_ids is not None:
            sample_ids = sample_ids[good_inds]

        return sample_ids, None, keep_inds, good_inds

    sample_ids, label_ids = _get_patch_ids(samples, patches_field)

    if index_label_ids is None:
        return sample_ids, label_ids, None, None

    keep_inds, good_inds, bad_ids = _parse_ids(
        label_ids,
        index_label_ids,
        "labels",
        allow_missing,
        warn_missing,
    )

    if bad_ids is not None:
        sample_ids = sample_ids[good_inds]
        label_ids = label_ids[good_inds]

    return sample_ids, label_ids, keep_inds, good_inds


def _get_patch_ids(
    samples, patches_field, handle_missing="skip", ref_sample_ids=None
):
    if samples._is_patches:
        sample_id_path = "sample_id"
    else:
        sample_id_path = "id"

    label_type, label_id_path = samples._get_label_field_path(
        patches_field, "id"
    )
    is_list_field = issubclass(label_type, fol._LABEL_LIST_FIELDS)

    sample_ids, label_ids = samples.values([sample_id_path, label_id_path])

    if ref_sample_ids is not None:
        sample_ids, label_ids = _apply_ref_sample_ids(
            sample_ids, label_ids, ref_sample_ids
        )

    if is_list_field:
        sample_ids, label_ids = _flatten_list_ids(
            sample_ids, label_ids, handle_missing
        )

    return np.array(sample_ids), np.array(label_ids)


def _apply_ref_sample_ids(sample_ids, label_ids, ref_sample_ids):
    ref_label_ids = [None] * len(ref_sample_ids)
    inds_map = {_id: i for i, _id in enumerate(ref_sample_ids)}
    for _id, _lid in zip(sample_ids, label_ids):
        idx = inds_map.get(_id, None)
        if idx is not None:
            ref_label_ids[idx] = _lid

    return ref_sample_ids, ref_label_ids


def _flatten_list_ids(sample_ids, label_ids, handle_missing):
    _sample_ids = []
    _label_ids = []
    _add_missing = handle_missing == "image"

    for _id, _lids in zip(sample_ids, label_ids):
        if _lids:
            for _lid in _lids:
                _sample_ids.append(_id)
                _label_ids.append(_lid)
        elif _add_missing:
            _sample_ids.append(_id)
            _label_ids.append(None)

    return _sample_ids, _label_ids


def _parse_ids(ids, index_ids, ftype, allow_missing, warn_missing):
    if np.array_equal(ids, index_ids):
        return None, None, None

    inds_map = {_id: idx for idx, _id in enumerate(index_ids)}

    keep_inds = []
    bad_inds = []
    bad_ids = []
    for _idx, _id in enumerate(ids):
        ind = inds_map.get(_id, None)
        if ind is not None:
            keep_inds.append(ind)
        else:
            bad_inds.append(_idx)
            bad_ids.append(_id)

    num_missing_index = len(index_ids) - len(keep_inds)
    if num_missing_index > 0:
        if not allow_missing:
            raise ValueError(
                "The index contains %d %s that are not present in the "
                "provided collection" % (num_missing_index, ftype)
            )

        if warn_missing:
            logger.warning(
                "Ignoring %d %s from the index that are not present in the "
                "provided collection",
                num_missing_index,
                ftype,
            )

    num_missing_collection = len(bad_ids)
    if num_missing_collection > 0:
        if not allow_missing:
            raise ValueError(
                "The provided collection contains %d %s not present in the "
                "index" % (num_missing_collection, ftype)
            )

        if warn_missing:
            logger.warning(
                "Ignoring %d %s from the provided collection that are not "
                "present in the index",
                num_missing_collection,
                ftype,
            )

        bad_inds = np.array(bad_inds, dtype=np.int64)

        good_inds = np.full(ids.shape, True)
        good_inds[bad_inds] = False
    else:
        good_inds = None
        bad_ids = None

    keep_inds = np.array(keep_inds, dtype=np.int64)

    return keep_inds, good_inds, bad_ids


def skip_ids(samples, ids, patches_field=None, warn_existing=False):
    sample_ids, label_ids = get_ids(samples, patches_field=patches_field)

    if patches_field is not None:
        exclude_ids = list(set(label_ids) - set(ids))
        num_existing = len(exclude_ids)

        if num_existing > 0:
            if warn_existing:
                logger.warning("Skipping %d existing label IDs", num_existing)

            samples = samples.exclude_labels(
                ids=exclude_ids, fields=patches_field
            )
    else:
        exclude_ids = list(set(sample_ids) - set(ids))
        num_existing = len(exclude_ids)

        if num_existing > 0:
            if warn_existing:
                logger.warning("Skipping %d existing sample IDs", num_existing)

        samples = samples.exclude(exclude_ids)

    return samples


def add_ids(
    sample_ids,
    label_ids,
    index_sample_ids,
    index_label_ids,
    patches_field=None,
    overwrite=True,
    allow_existing=True,
    warn_existing=False,
):
    if patches_field is not None:
        ids = label_ids
        index_ids = index_label_ids
    else:
        ids = sample_ids
        index_ids = index_sample_ids

    ii = []
    jj = []

    ids_map = {_id: _i for _i, _id in enumerate(index_ids)}
    new_idx = len(index_ids)
    for _i, _id in enumerate(ids):
        _idx = ids_map.get(_id, None)
        if _idx is None:
            ii.append(_i)
            jj.append(new_idx)
            new_idx += 1
        elif overwrite:
            ii.append(_i)
            jj.append(_idx)

    ii = np.array(ii)
    jj = np.array(jj)

    n = len(index_sample_ids)

    if not allow_existing:
        existing_inds = np.nonzero(jj < n)[0]
        num_existing = existing_inds.size

        if num_existing > 0:
            if warn_existing:
                logger.warning(
                    "Ignoring %d IDs (eg '%s') that are already present in "
                    "the index",
                    num_existing,
                    ids[ii[0]],
                )

                ii = np.delete(ii, existing_inds)
                jj = np.delete(jj, existing_inds)
            else:
                raise ValueError(
                    "Found %d IDs (eg '%s') that are already present in the "
                    "index" % (num_existing, ids[ii[0]])
                )

    if ii.size > 0:
        sample_ids = np.array(sample_ids)
        if patches_field is not None:
            label_ids = np.array(label_ids)

        m = jj[-1] - n + 1

        if n == 0:
            index_sample_ids = np.array([], dtype=sample_ids.dtype)
            if patches_field is not None:
                index_label_ids = np.array([], dtype=label_ids.dtype)

        if m > 0:
            index_sample_ids = np.concatenate(
                (index_sample_ids, np.empty(m, dtype=index_sample_ids.dtype))
            )
            if patches_field is not None:
                index_label_ids = np.concatenate(
                    (index_label_ids, np.empty(m, dtype=index_label_ids.dtype))
                )

        index_sample_ids[jj] = sample_ids[ii]
        if patches_field is not None:
            index_label_ids[jj] = label_ids[ii]

    return index_sample_ids, index_label_ids, ii, jj


def add_embeddings(
    samples,
    embeddings,
    sample_ids,
    label_ids,
    embeddings_field,
    patches_field=None,
):
    dataset = samples._dataset

    if patches_field is not None:
        _, embeddings_path = dataset._get_label_field_path(
            patches_field, embeddings_field
        )

        values = dict(zip(label_ids, embeddings))
        dataset.set_label_values(embeddings_path, values)
    else:
        values = dict(zip(sample_ids, embeddings))
        dataset.set_values(embeddings_field, values, key_field="id")


def remove_ids(
    sample_ids,
    label_ids,
    index_sample_ids,
    index_label_ids,
    patches_field=None,
    allow_missing=True,
    warn_missing=False,
):
    rm_inds = []

    if sample_ids is not None:
        rm_inds.extend(
            _find_ids(
                sample_ids,
                index_sample_ids,
                allow_missing,
                warn_missing,
                "sample",
            )
        )

    if label_ids is not None:
        rm_inds.extend(
            _find_ids(
                label_ids,
                index_label_ids,
                allow_missing,
                warn_missing,
                "label",
            )
        )

    rm_inds = np.array(rm_inds)

    if rm_inds.size > 0:
        index_sample_ids = np.delete(index_sample_ids, rm_inds)
        if patches_field is not None:
            index_label_ids = np.delete(index_label_ids, rm_inds)

    return index_sample_ids, index_label_ids, rm_inds


def _find_ids(ids, index_ids, allow_missing, warn_missing, ftype):
    found_inds = []
    missing_ids = []

    ids_map = {_id: _i for _i, _id in enumerate(index_ids)}
    for _id in ids:
        ind = ids_map.get(_id, None)
        if ind is not None:
            found_inds.append(ind)
        elif not allow_missing:
            missing_ids.append(_id)

    num_missing = len(missing_ids)

    if num_missing > 0:
        if not allow_missing:
            raise ValueError(
                "Found %d %d IDs (eg '%s') that are not present in the index"
                % (num_missing, ftype, missing_ids[0])
            )

        if warn_missing:
            logger.warning(
                "Ignoring %d %d IDs (eg '%s') that are not present in the "
                "index",
                num_missing,
                ftype,
                missing_ids[0],
            )

    return found_inds


def remove_embeddings(
    samples,
    embeddings_field,
    sample_ids=None,
    label_ids=None,
    patches_field=None,
):
    dataset = samples._dataset

    if patches_field is not None:
        _, embeddings_path = dataset._get_label_field_path(
            patches_field, embeddings_field
        )

        if sample_ids is not None and label_ids is None:
            _, id_path = dataset._get_label_field_path(patches_field, "id")
            label_ids = dataset.select(sample_ids).values(id_path, unwind=True)

        if label_ids is not None:
            values = dict(zip(label_ids, itertools.repeat(None)))
            dataset.set_label_values(embeddings_path, values)
    elif sample_ids is not None:
        values = dict(zip(sample_ids, itertools.repeat(None)))
        dataset.set_values(embeddings_field, values, key_field="id")


def filter_values(values, keep_inds, patches_field=None):
    if patches_field:
        _values = list(itertools.chain.from_iterable(values))
    else:
        _values = values

    _values = np.asarray(_values)

    if _values.size == keep_inds.size:
        _values = _values[keep_inds]
    else:
        num_expected = np.count_nonzero(keep_inds)
        if _values.size != num_expected:
            raise ValueError(
                "Expected %d raw values or %d pre-filtered values; found %d "
                "values" % (keep_inds.size, num_expected, values.size)
            )

    # @todo we might need to re-ravel patch values here in the future
    # We currently do not do this because all downstream users of this data
    # will gracefully handle either flat or nested list data

    return _values


def get_values(samples, path_or_expr, ids, patches_field=None):
    _validate_args(samples, path_or_expr, patches_field)
    return samples._get_values_by_id(
        path_or_expr, ids, link_field=patches_field
    )


def parse_embeddings_field(
    samples, embeddings_field, patches_field=None, allow_embedded=True
):
    if not etau.is_str(embeddings_field):
        raise ValueError(
            "Invalid embeddings_field=%s; expected a string field name"
            % embeddings_field
        )

    if patches_field is None:
        _embeddings_field, is_frame_field = samples._handle_frame_field(
            embeddings_field
        )

        if not allow_embedded and "." in _embeddings_field:
            ftype = "frame" if is_frame_field else "sample"
            raise ValueError(
                "Invalid embeddings_field=%s; expected a top-level %s field "
                "name that contains no '.'" % (_embeddings_field, ftype)
            )

        return embeddings_field

    if embeddings_field.startswith(patches_field + "."):
        _, root = samples._get_label_field_path(patches_field) + "."
        if not embeddings_field.startswith(root):
            raise ValueError(
                "Invalid embeddings_field=%s for patches_field=%s"
                % (embeddings_field, patches_field)
            )

        embeddings_field = embeddings_field[len(root) + 1]

    if not allow_embedded and "." in embeddings_field:
        raise ValueError(
            "Invalid embeddings_field=%s for patches_field=%s; expected a "
            "label attribute name that contains no '.'"
            % (embeddings_field, patches_field)
        )

    return embeddings_field


def get_embeddings(
    samples,
    model=None,
    patches_field=None,
    embeddings_field=None,
    embeddings=None,
    force_square=False,
    alpha=None,
    handle_missing="skip",
    agg_fcn=None,
    batch_size=None,
    num_workers=None,
    skip_failures=True,
):
    if model is None and embeddings_field is None and embeddings is None:
        return _empty_embeddings(patches_field)

    if isinstance(embeddings, fob.SimilarityIndex):
        allow_missing = handle_missing == "skip"
        return get_embeddings_from_index(
            samples,
            embeddings,
            patches_field=patches_field,
            allow_missing=allow_missing,
            warn_missing=True,
        )

    if embeddings is None and model is not None:
        if etau.is_str(model):
            model = foz.load_zoo_model(model)

        if patches_field is not None:
            logger.info("Computing patch embeddings...")
            embeddings = samples.compute_patch_embeddings(
                model,
                patches_field,
                embeddings_field=embeddings_field,
                force_square=force_square,
                alpha=alpha,
                handle_missing=handle_missing,
                batch_size=batch_size,
                num_workers=num_workers,
                skip_failures=skip_failures,
            )
        else:
            logger.info("Computing embeddings...")
            embeddings = samples.compute_embeddings(
                model,
                embeddings_field=embeddings_field,
                batch_size=batch_size,
                num_workers=num_workers,
                skip_failures=skip_failures,
            )

    if embeddings is None and embeddings_field is not None:
        embeddings, samples = _load_embeddings(
            samples, embeddings_field, patches_field=patches_field
        )
        ref_sample_ids = None
    else:
        if isinstance(embeddings, dict):
            embeddings = [
                embeddings.get(_id, None) for _id in samples.values("id")
            ]

        embeddings, ref_sample_ids = _handle_missing_embeddings(
            embeddings, samples
        )

    if not isinstance(embeddings, np.ndarray) and not embeddings:
        return _empty_embeddings(patches_field)

    if patches_field is not None:
        if agg_fcn is not None:
            embeddings = np.stack([agg_fcn(e) for e in embeddings])
        else:
            embeddings = np.concatenate(embeddings, axis=0)
    elif not isinstance(embeddings, np.ndarray):
        embeddings = np.stack(embeddings)

    if agg_fcn is not None:
        patches_field = None

    sample_ids, label_ids = get_ids(
        samples,
        patches_field=patches_field,
        data=embeddings,
        data_type="embeddings",
        handle_missing=handle_missing,
        ref_sample_ids=ref_sample_ids,
    )

    return embeddings, sample_ids, label_ids


def get_unique_name(name, ref_names):
    ref_names = set(ref_names)

    if name in ref_names:
        name += "-" + _get_random_characters(6)

    while name in ref_names:
        name += _get_random_characters(1)

    return name


def _get_random_characters(n):
    return "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(n)
    )


def _empty_embeddings(patches_field):
    embeddings = np.empty((0, 0), dtype=float)
    sample_ids = np.array([], dtype="<U24")

    if patches_field is not None:
        label_ids = np.array([], dtype="<U24")
    else:
        label_ids = None

    return embeddings, sample_ids, label_ids


def _load_embeddings(samples, embeddings_field, patches_field=None):
    if patches_field is not None:
        label_type, embeddings_path = samples._get_label_field_path(
            patches_field, embeddings_field
        )
        is_list_field = issubclass(label_type, fol._LABEL_LIST_FIELDS)
    else:
        embeddings_path = embeddings_field
        is_list_field = False

    if is_list_field:
        samples = samples.filter_labels(
            patches_field, F(embeddings_field) != None
        )
    else:
        samples = samples.match(F(embeddings_path) != None)

    if samples.has_field(embeddings_path):
        _field = None
    else:
        _field = fof.VectorField()

    embeddings = samples.values(embeddings_path, _field=_field)

    if is_list_field:
        embeddings = [np.stack(e) for e in embeddings if e]

    return embeddings, samples


def _validate_args(samples, path_or_expr, patches_field):
    if patches_field is not None:
        _validate_patches_args(samples, path_or_expr, patches_field)
    else:
        _validate_samples_args(samples, path_or_expr)


def _validate_samples_args(samples, path_or_expr):
    if not etau.is_str(path_or_expr):
        return

    path, _, list_fields, _, _ = samples._parse_field_name(path_or_expr)

    if list_fields:
        raise ValueError(
            "Values path '%s' contains invalid list field '%s'"
            % (path, list_fields[0])
        )


def _validate_patches_args(samples, path_or_expr, patches_field):
    if etau.is_str(path_or_expr) and not path_or_expr.startswith(
        patches_field + "."
    ):
        raise ValueError(
            "Values path '%s' must start with patches field '%s'"
            % (path_or_expr, patches_field)
        )

    if (
        isinstance(samples, fop.PatchesView)
        and patches_field != samples.patches_field
    ):
        raise ValueError(
            "This patches view contains labels from field '%s', not "
            "'%s'" % (samples.patches_field, patches_field)
        )

    if isinstance(
        samples, fop.EvaluationPatchesView
    ) and patches_field not in (
        samples.gt_field,
        samples.pred_field,
    ):
        raise ValueError(
            "This evaluation patches view contains patches from "
            "fields '%s' and '%s', not '%s'"
            % (samples.gt_field, samples.pred_field, patches_field)
        )


def _handle_missing_embeddings(embeddings, samples):
    if isinstance(embeddings, np.ndarray):
        return embeddings, None

    missing_inds = []
    for idx, embedding in enumerate(embeddings):
        if embedding is None:
            missing_inds.append(idx)

    if not missing_inds:
        return embeddings, None

    embeddings = [e for e in embeddings if e is not None]
    ref_sample_ids = list(np.delete(samples.values("id"), missing_inds))

    return embeddings, ref_sample_ids



"""
Script to generate a screenshot of near-duplicates using the uniqueness and
ranking.

You need to run this inside of an interactive shell like IPython because the
system will launch the dashboard and we need to have that process independent
of the python code.

Outputs are placed in the `outputs/` directory.

Usage::

    # From inside IPython
    run shot_neardups.py

Copyright 2017-2023, Voxel51, Inc.
voxel51.com
"""
import os
import time

import eta.core.utils as etau

import fiftyone as fo
import fiftyone.zoo as foz

import fiftyone.brain as fob


OUTPUT_DIR = "outputs"


print("Working on near-duplicates in CIFAR-10 Test")

dataset = foz.load_zoo_dataset("cifar10", split="test")

fob.compute_uniqueness(dataset)

view = dataset.view().sort_by("uniqueness")

print("Launching dashboard...")
session = fo.launch_dashboard(view=view)

input("Press enter when you are ready for the screenshot:")

localtime = lambda: time.strftime("%Y%m%d-%H%M%S", time.localtime())
outname = "neardups_cifar10_test" + localtime() + ".png"
outpath = os.path.join(OUTPUT_DIR, outname)

etau.save_window_snapshot("FiftyOne", outpath)



"""
Similarity tests.

Usage::

    pytest tests/intensive/test_similarity.py -s -k test_XXX

Qdrant setup::

    docker pull qdrant/qdrant
    docker run -p 6333:6333

    pip install qdrant-client

Pinecone setup::

    # Sign up at https://www.pinecone.io
    # Download API key and environment

    pip install pinecone-client

Brain config setup at `~/.fiftyone/brain_config.json`::

    {
        "similarity_backends": {
            "pinecone": {
                "environment": "us-east-1-aws",
                "api_key": "XXXXXXXX"
            },
            "qdrant": {
                "url": "http://localhost:6333"
            }
        }
    }

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import random
import unittest

import numpy as np

import fiftyone as fo
import fiftyone.brain as fob  # pylint: disable=import-error,no-name-in-module
import fiftyone.zoo as foz
from fiftyone import ViewField as F


def test_brain_config():
    similarity_backends = fob.brain_config.similarity_backends

    assert "sklearn" in similarity_backends

    assert "qdrant" in similarity_backends
    assert "url" in similarity_backends["qdrant"]

    assert "pinecone" in similarity_backends
    assert "api_key" in similarity_backends["pinecone"]
    assert "environment" in similarity_backends["pinecone"]


def test_image_similarity_backends():
    dataset = foz.load_zoo_dataset("quickstart")

    index1 = fob.compute_similarity(
        dataset,
        model="clip-vit-base32-torch",
        metric="euclidean",
        embeddings=False,
        backend="sklearn",
        brain_key="clip_sklearn",
    )

    index2 = fob.compute_similarity(
        dataset,
        model="clip-vit-base32-torch",
        metric="euclidean",
        embeddings=False,
        backend="qdrant",
        brain_key="clip_qdrant",
    )

    index3 = fob.compute_similarity(
        dataset,
        model="clip-vit-base32-torch",
        metric="euclidean",
        embeddings=False,
        backend="pinecone",
        brain_key="clip_pinecone",
    )

    embeddings, sample_ids, _ = index1.compute_embeddings(dataset)

    index1.add_to_index(embeddings, sample_ids)
    index1.save()
    index1.reload()
    assert index1.total_index_size == 200
    assert index1.index_size == 200
    assert index1.missing_size is None

    index2.add_to_index(embeddings, sample_ids)
    assert index2.total_index_size == 200
    assert index2.index_size == 200
    assert index2.missing_size is None

    index3.add_to_index(embeddings, sample_ids)
    assert index3.total_index_size == 200
    assert index3.index_size == 200
    assert index3.missing_size is None

    prompt = "kites high in the air"

    view1 = dataset.sort_by_similarity(prompt, k=10, brain_key="clip_sklearn")
    assert len(view1) == 10

    view2 = dataset.sort_by_similarity(prompt, k=10, brain_key="clip_qdrant")
    assert len(view2) == 10

    view3 = dataset.sort_by_similarity(prompt, k=10, brain_key="clip_pinecone")
    assert len(view3) == 10

    del index1
    del index2
    del index3
    dataset.clear_cache()

    print(dataset.get_brain_info("clip_sklearn"))
    print(dataset.get_brain_info("clip_qdrant"))
    print(dataset.get_brain_info("clip_pinecone"))

    index1 = dataset.load_brain_results("clip_sklearn")
    index2 = dataset.load_brain_results("clip_qdrant")
    index3 = dataset.load_brain_results("clip_pinecone")

    assert index1.total_index_size == 200
    assert index2.total_index_size == 200
    assert index3.total_index_size == 200

    embeddings1, sample_ids1, _ = index1.get_embeddings()
    assert embeddings1.shape == (200, 512)
    assert sample_ids1.shape == (200,)

    embeddings2, sample_ids2, _ = index2.get_embeddings()
    assert embeddings2.shape == (200, 512)
    assert sample_ids2.shape == (200,)

    # Pinecone requires IDs
    # embeddings3, sample_ids3, _ = index3.get_embeddings()
    # assert embeddings3.shape == (200, 512)
    # assert sample_ids3.shape == (200,)

    ids = random.sample(list(index1.sample_ids), 100)

    embeddings1, sample_ids1, _ = index1.get_embeddings(sample_ids=ids)
    embeddings2, sample_ids2, _ = index2.get_embeddings(sample_ids=ids)
    embeddings3, sample_ids3, _ = index3.get_embeddings(sample_ids=ids)

    embeddings2_dict = dict(zip(sample_ids2, embeddings2))
    embeddings3_dict = dict(zip(sample_ids3, embeddings3))

    _embeddings2 = np.array([embeddings2_dict[i] for i in sample_ids1])
    _embeddings3 = np.array([embeddings3_dict[i] for i in sample_ids1])

    assert embeddings1.shape == (100, 512)
    assert sample_ids1.shape == (100,)

    assert embeddings2.shape == (100, 512)
    assert sample_ids2.shape == (100,)

    assert embeddings3.shape == (100, 512)
    assert sample_ids3.shape == (100,)

    assert set(sample_ids1) == set(sample_ids2)
    assert set(sample_ids1) == set(sample_ids3)

    assert np.allclose(embeddings1, _embeddings2)
    assert np.allclose(embeddings1, _embeddings3)

    index1.remove_from_index(sample_ids=ids)
    index2.remove_from_index(sample_ids=ids)
    index3.remove_from_index(sample_ids=ids)

    assert index1.total_index_size == 100
    assert index2.total_index_size == 100
    assert index3.total_index_size == 100

    index1.cleanup()
    index2.cleanup()
    index3.cleanup()

    dataset.delete_brain_run("clip_sklearn")
    dataset.delete_brain_run("clip_qdrant")
    dataset.delete_brain_run("clip_pinecone")
    dataset.delete()


def test_patch_similarity_backends():
    dataset = foz.load_zoo_dataset("quickstart")

    index1 = fob.compute_similarity(
        dataset,
        patches_field="ground_truth",
        model="clip-vit-base32-torch",
        metric="euclidean",
        embeddings=False,
        backend="sklearn",
        brain_key="gt_clip_sklearn",
    )

    index2 = fob.compute_similarity(
        dataset,
        patches_field="ground_truth",
        model="clip-vit-base32-torch",
        metric="euclidean",
        embeddings=False,
        backend="qdrant",
        brain_key="gt_clip_qdrant",
    )

    index3 = fob.compute_similarity(
        dataset,
        patches_field="ground_truth",
        model="clip-vit-base32-torch",
        metric="euclidean",
        embeddings=False,
        backend="pinecone",
        brain_key="gt_clip_pinecone",
    )

    embeddings, sample_ids, label_ids = index1.compute_embeddings(dataset)

    index1.add_to_index(embeddings, sample_ids, label_ids=label_ids)
    index1.save()
    index1.reload()
    assert index1.total_index_size == 1232
    assert index1.index_size == 1232
    assert index1.missing_size is None

    index2.add_to_index(embeddings, sample_ids, label_ids=label_ids)
    assert index2.total_index_size == 1232
    assert index2.index_size == 1232
    assert index2.missing_size is None

    index3.add_to_index(embeddings, sample_ids, label_ids=label_ids)
    assert index3.total_index_size == 1232
    assert index3.index_size == 1232
    assert index3.missing_size is None

    view = dataset.to_patches("ground_truth")

    prompt = "cute puppies"

    view1 = view.sort_by_similarity(prompt, k=10, brain_key="gt_clip_sklearn")
    assert len(view1) == 10

    view2 = view.sort_by_similarity(prompt, k=10, brain_key="gt_clip_qdrant")
    assert len(view2) == 10

    view3 = view.sort_by_similarity(prompt, k=10, brain_key="gt_clip_pinecone")
    assert len(view3) == 10

    del index1
    del index2
    del index3
    dataset.clear_cache()

    print(dataset.get_brain_info("gt_clip_sklearn"))
    print(dataset.get_brain_info("gt_clip_qdrant"))
    print(dataset.get_brain_info("gt_clip_pinecone"))

    index1 = dataset.load_brain_results("gt_clip_sklearn")
    index2 = dataset.load_brain_results("gt_clip_qdrant")
    index3 = dataset.load_brain_results("gt_clip_pinecone")

    assert index1.total_index_size == 1232
    assert index2.total_index_size == 1232
    assert index3.total_index_size == 1232

    embeddings1, sample_ids1, label_ids1 = index1.get_embeddings()
    assert embeddings1.shape == (1232, 512)
    assert sample_ids1.shape == (1232,)
    assert label_ids1.shape == (1232,)

    embeddings2, sample_ids2, label_ids2 = index2.get_embeddings()
    assert embeddings2.shape == (1232, 512)
    assert sample_ids2.shape == (1232,)
    assert label_ids2.shape == (1232,)

    # Pinecone requires IDs
    # embeddings3, sample_ids3, label_ids3 = index3.get_embeddings()
    # assert embeddings3.shape == (1232, 512)
    # assert sample_ids3.shape == (1232,)
    # assert label_ids3.shape == (1232,)

    ids = random.sample(list(index1.label_ids), 100)

    embeddings1, sample_ids1, label_ids1 = index1.get_embeddings(label_ids=ids)
    embeddings2, sample_ids2, label_ids2 = index2.get_embeddings(label_ids=ids)
    embeddings3, sample_ids3, label_ids3 = index3.get_embeddings(label_ids=ids)

    embeddings2_dict = dict(zip(label_ids2, embeddings2))
    embeddings3_dict = dict(zip(label_ids3, embeddings3))

    _embeddings2 = np.array([embeddings2_dict[i] for i in label_ids1])
    _embeddings3 = np.array([embeddings3_dict[i] for i in label_ids1])

    assert embeddings1.shape == (100, 512)
    assert sample_ids1.shape == (100,)
    assert label_ids1.shape == (100,)

    assert embeddings2.shape == (100, 512)
    assert sample_ids2.shape == (100,)
    assert label_ids2.shape == (100,)

    assert embeddings3.shape == (100, 512)
    assert sample_ids3.shape == (100,)
    assert label_ids3.shape == (100,)

    assert set(label_ids1) == set(label_ids2)
    assert set(label_ids1) == set(label_ids3)

    assert np.allclose(embeddings1, _embeddings2)
    assert np.allclose(embeddings1, _embeddings3)

    index1.remove_from_index(label_ids=ids)
    index2.remove_from_index(label_ids=ids)
    index3.remove_from_index(label_ids=ids)

    assert index1.total_index_size == 1132
    assert index2.total_index_size == 1132
    assert index3.total_index_size == 1132

    index1.cleanup()
    index2.cleanup()
    index3.cleanup()

    dataset.delete_brain_run("gt_clip_sklearn")
    dataset.delete_brain_run("gt_clip_qdrant")
    dataset.delete_brain_run("gt_clip_pinecone")


def test_images():
    dataset = _load_images_dataset()

    index = dataset.load_brain_results("img_sim")

    assert index.total_index_size == len(dataset)
    assert set(dataset.values("id")) == set(index.sample_ids)


def test_images_subset():
    dataset = _load_images_dataset()

    index = dataset.load_brain_results("img_sim")

    view = dataset.take(10)
    index.use_view(view)

    assert index.index_size == len(view)
    assert set(view.values("id")) == set(index.current_sample_ids)


def test_images_missing():
    dataset = _load_images_dataset().limit(4).clone()
    dataset.add_samples(
        [
            fo.Sample(filepath="non-existent1.png"),
            fo.Sample(filepath="non-existent2.png"),
            fo.Sample(filepath="non-existent3.png"),
            fo.Sample(filepath="non-existent4.png"),
        ]
    )

    sample_ids = dataset[:4].values("id")

    index = fob.compute_similarity(dataset, batch_size=1)

    assert index.total_index_size == 4
    assert set(sample_ids) == set(index.sample_ids)

    model = foz.load_zoo_model("inception-v3-imagenet-torch")
    index = fob.compute_similarity(
        dataset,
        model=model,
        embeddings="embeddings_missing",
        batch_size=1,
    )

    assert len(dataset.exists("embeddings_missing")) == 4
    assert index.index_size == 4
    assert set(sample_ids) == set(index.sample_ids)


def test_patches():
    dataset = _load_patches_dataset()

    index = dataset.load_brain_results("gt_sim")

    label_ids = dataset.values("ground_truth.detections.id", unwind=True)

    assert index.total_index_size == len(label_ids)
    assert set(label_ids) == set(index.label_ids)


def test_patches_subset():
    dataset = _load_patches_dataset()

    index = dataset.load_brain_results("gt_sim")

    label_ids = dataset.values("ground_truth.detections.id", unwind=True)

    assert index.total_index_size == len(label_ids)
    assert set(label_ids) == set(index.label_ids)

    view = dataset.filter_labels("ground_truth", F("label") == "person")
    index.use_view(view)

    label_ids = view.values("ground_truth.detections.id", unwind=True)

    assert index.index_size == len(label_ids)
    assert set(label_ids) == set(index.current_label_ids)


def test_patches_missing():
    dataset = _load_patches_dataset().limit(4).clone()
    dataset.add_samples(
        [
            fo.Sample(filepath="non-existent1.png"),
            fo.Sample(filepath="non-existent2.png"),
            fo.Sample(filepath="non-existent3.png"),
            fo.Sample(filepath="non-existent4.png"),
        ]
    )

    for sample in dataset[4:]:
        sample["ground_truth"] = fo.Detections(
            detections=[fo.Detection(bounding_box=[0.1, 0.1, 0.8, 0.8])]
        )
        sample.save()

    index = fob.compute_similarity(
        dataset, patches_field="ground_truth", batch_size=1
    )

    num_patches = dataset[:4].count("ground_truth.detections")
    label_ids = dataset[:4].values("ground_truth.detections.id", unwind=True)

    assert index.total_index_size == num_patches
    assert set(label_ids) == set(index.label_ids)

    model = foz.load_zoo_model("inception-v3-imagenet-torch")
    index = fob.compute_similarity(
        dataset,
        model=model,
        patches_field="ground_truth",
        embeddings="embeddings_missing",
        batch_size=1,
    )

    view = dataset.filter_labels(
        "ground_truth", F("embeddings_missing") != None
    )

    assert view.count("ground_truth.detections") == num_patches
    assert index.total_index_size == num_patches
    assert set(label_ids) == set(index.label_ids)


def _load_images_dataset():
    name = "test-similarity-images"

    if fo.dataset_exists(name):
        return fo.load_dataset(name)

    return _make_images_dataset(name)


def _load_patches_dataset():
    name = "test-similarity-patches"

    if fo.dataset_exists(name):
        return fo.load_dataset(name)

    return _make_patches_dataset(name)


def _make_images_dataset(name):
    dataset = foz.load_zoo_dataset(
        "quickstart", max_samples=20, dataset_name=name
    )
    model = foz.load_zoo_model("inception-v3-imagenet-torch")

    # Embed images
    dataset.compute_embeddings(
        model, embeddings_field="embeddings", batch_size=8
    )

    # Image similarity
    fob.compute_similarity(
        dataset, embeddings="embeddings", brain_key="img_sim"
    )

    return dataset


def _make_patches_dataset(name):
    dataset = foz.load_zoo_dataset(
        "quickstart", max_samples=20, dataset_name=name
    )
    model = foz.load_zoo_model("inception-v3-imagenet-torch")

    # Embed ground truth patches
    dataset.compute_patch_embeddings(
        model,
        "ground_truth",
        embeddings_field="embeddings",
        batch_size=8,
        force_square=True,
    )

    # Patch similarity
    fob.compute_similarity(
        dataset,
        patches_field="ground_truth",
        embeddings="embeddings",
        brain_key="gt_sim",
    )

    return dataset


if __name__ == "__main__":
    fo.config.show_progress_bars = True
    unittest.main(verbosity=2)



"""
Brain models.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from copy import deepcopy
import logging
import os

from eta.core.config import ConfigError
import eta.core.learning as etal
import eta.core.models as etam

import fiftyone.core.models as fom


logger = logging.getLogger(__name__)


_THIS_DIR = os.path.dirname(os.path.abspath(__file__))
_MODELS_MANIFEST_PATH = os.path.join(_THIS_DIR, "manifest.json")
_MODELS_DIR = os.path.join(_THIS_DIR, "cache")


def list_models():
    """Returns the list of available models.

    Returns:
        a list of model names
    """
    manifest = _load_models_manifest()
    return sorted([model.name for model in manifest])


def list_downloaded_models():
    """Returns information about the models that have been downloaded.

    Returns:
        a dict mapping model names to (model path, ``eta.core.models.Model``)
        tuples
    """
    manifest = _load_models_manifest()
    models = {}
    for model in manifest:
        if model.is_in_dir(_MODELS_DIR):
            model_path = model.get_path_in_dir(_MODELS_DIR)
            models[model.name] = (model_path, model)

    return models


def is_model_downloaded(name):
    """Determines whether the model of the given name is downloaded.

    Args:
        name: the name of the model, which can have ``@<ver>`` appended to
            refer to a specific version of the model. If no version is
            specified, the latest version of the model is used

    Returns:
        True/False
    """
    model = _get_model(name)
    return model.is_in_dir(_MODELS_DIR)


def download_model(name, overwrite=False):
    """Downloads the model of the given name.

    If the model is already downloaded, it is not re-downloaded unless
    ``overwrite == True`` is specified.

    Args:
        name: the name of the model, which can have ``@<ver>`` appended to
            refer to a specific version of the model. If no version is
            specified, the latest version of the model is used. Call
            :func:`list_models` to see the available models
        overwrite (False): whether to overwrite any existing files

    Returns:
        tuple of

        -   model: the ``eta.core.models.Model`` instance for the model
        -   model_path: the path to the downloaded model on disk
    """
    model, model_path = _get_model_in_dir(name)

    if not overwrite and is_model_downloaded(name):
        logger.info("Model '%s' is already downloaded", name)
    else:
        model.manager.download_model(model_path, force=overwrite)

    return model, model_path


def install_model_requirements(name, error_level=0):
    """Installs any package requirements for the model with the given name.

    Args:
        name: the name of the model, which can have ``@<ver>`` appended to
            refer to a specific version of the model. If no version is
            specified, the latest version of the model is used. Call
            :func:`list_models` to see the available models
        error_level: the error level to use, defined as:

            0: raise error if a requirement install fails
            1: log warning if a requirement install fails
            2: ignore install fails
    """
    model = _get_model(name)
    model.install_requirements(error_level=error_level)


def ensure_model_requirements(name, error_level=0):
    """Ensures that the package requirements for the model with the given name
    are satisfied.

    Args:
        name: the name of the model, which can have ``@<ver>`` appended to
            refer to a specific version of the model. If no version is
            specified, the latest version of the model is used. Call
            :func:`list_models` to see the available models
        error_level: the error level to use, defined as:

            0: raise error if a requirement is not satisfied
            1: log warning if a requirement is not satisifed
            2: ignore unsatisifed requirements
    """
    model = _get_model(name)
    model.ensure_requirements(error_level=error_level)


def load_model(
    name,
    download_if_necessary=True,
    install_requirements=False,
    error_level=0,
    **kwargs
):
    """Loads the model of the given name.

    By default, the model will be downloaded if necessary.

    Args:
        name: the name of the model, which can have ``@<ver>`` appended to
            refer to a specific version of the model. If no version is
            specified, the latest version of the model is used. Call
            :func:`list_models` to see the available models
        download_if_necessary (True): whether to download the model if it is
            not found in the specified directory
        install_requirements: whether to install any requirements before
            loading the model. By default, this is False
        error_level: the error level to use, defined as:

            0: raise error if a requirement is not satisfied
            1: log warning if a requirement is not satisifed
            2: ignore unsatisifed requirements

        **kwargs: keyword arguments to inject into the model's ``Config``
            instance

    Returns:
        a :class:`fiftyone.core.models.Model`
    """
    model = _get_model(name)

    if not model.is_in_dir(_MODELS_DIR):
        if not download_if_necessary:
            raise ValueError("Model '%s' is not downloaded" % name)

        download_model(name)

    if install_requirements:
        model.install_requirements(error_level=error_level)
    else:
        model.ensure_requirements(error_level=error_level)

    config_dict = deepcopy(model.default_deployment_config_dict)
    model_path = model.get_path_in_dir(_MODELS_DIR)

    return fom.load_model(config_dict, model_path=model_path, **kwargs)


def find_model(name):
    """Returns the path to the model on disk.

    The model must be downloaded. Use :func:`download_model` to download
    models.

    Args:
        name: the name of the model, which can have ``@<ver>`` appended to
            refer to a specific version of the model. If no version is
            specified, the latest version of the model is used

    Returns:
        the path to the model on disk

    Raises:
        ValueError: if the model does not exist or has not been downloaded
    """
    model, model_path = _get_model_in_dir(name)
    if not model.is_model_downloaded(model_path):
        raise ValueError("Model '%s' is not downloaded" % name)

    return model_path


def get_model(name):
    """Returns the ``eta.core.models.Model`` instance for the model with the
    given name.

    Args:
        name: the name of the model

    Returnsn ``eta.core.models.Model``:class:`ZooModel`
    """
    return _get_model(name)


def delete_model(name):
    """Deletes the model from local disk, if necessary.

    Args:
        name: the name of the model, which can have ``@<ver>`` appended to
            refer to a specific version of the model. If no version is
            specified, the latest version of the model is used
    """
    model, model_path = _get_model_in_dir(name)
    model.flush_model(model_path)


class HasBrainModel(etal.HasPublishedModel):
    """Mixin class for Config classes of :class:`fiftyone.core.models.Model`
    instances whose models are stored privately by the FiftyOne Brain.
    """

    def download_model_if_necessary(self):
        # pylint: disable=attribute-defined-outside-init
        if not self.model_name and not self.model_path:
            raise ConfigError(
                "Either `model_name` or `model_path` must be provided"
            )

        if self.model_path is None:
            self.model_path = download_model(self.model_name)

    @classmethod
    def _get_model(cls, model_name):
        return get_model(model_name)


def _load_models_manifest():
    return etam.ModelsManifest.from_json(_MODELS_MANIFEST_PATH)


def _get_model_in_dir(name):
    model = _get_model(name)
    model_path = model.get_path_in_dir(_MODELS_DIR)
    return model, model_path


def _get_model(name):
    if etam.Model.has_version_str(name):
        return _get_exact_model(name)

    return _get_latest_model(name)


def _get_exact_model(name):
    manifest = _load_models_manifest()
    try:
        return manifest.get_model_with_name(name)
    except etam.ModelError:
        raise ValueError("No model with name '%s' was found" % name)


def _get_latest_model(base_name):
    manifest = _load_models_manifest()
    try:
        return manifest.get_latest_model_with_base_name(base_name)
    except etam.ModelError:
        raise ValueError("No models found with base name '%s'" % base_name)



"""
Script to plot a scalar value that was computed as certain points in an
iterative process.  It will average results from many such runs and plot shaded
error bars

It can also plot data-files that have multiple fields of results and select one
of them to plot per file.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""

import argparse
import json
import os

import matplotlib.pyplot as plt
import numpy as np


class Config:
    def __init__(self, d):
        '''d is a dictionary of arguments from command-line.'''
        self.colors = list(plt.get_cmap('Set1').colors)
        self.colors_N = len(self.colors)
        self.error_alpha = 0.2

        self.field = d['field']
        self.title = d['title']
        self.xlabel = d['xlabel']
        self.ylabel = d['ylabel']

    def color(self, index):
        '''Retrieve the color tuple for index with wrap-around.'''
        return self.colors[index % self.colors_N]


def read_to_array(config, json_path):
    '''Read the data in the file at json_path into a numpy array.

    The rows in the array are the experiment results and the columns are the
    values.  Returns a tuple with the data, indices where indices is a
    vector of the sample locations for the columns.
    '''
    with open(json_path, "rt") as f:
        d = json.load(f)

    num_rows = len(d.keys())

    assert num_rows >= 0

    indices = np.asarray(np.sort(list(d[next(iter(d.keys()))].keys())))
    ind_lut = {}
    for i, ind in enumerate(indices):
        ind_lut[ind] = i

    num_cols = len(indices)

    data = np.zeros((num_rows, num_cols))

    for i, (k, v) in enumerate(d.items()):
        # v is itself a dictionary with the keys being the indices and the
        # values being the actual graph values
        ind_check = np.sort(list(v.keys()))
        assert np.array_equal(ind_check, indices)

        if not isinstance(v[indices[0]], dict):
            # build the values this way in case they were not ordered in the dict
            for j, index in enumerate(indices):
                data[i][ind_lut[index]] = v[index]
        else:
            if not config.field:
                print(f'fail: data file has dicts but no field specified')
            for j, index in enumerate(indices):
                data[i][ind_lut[index]] = v[index][config.field]

    return indices, data


def main_plot(config, list_of_inputs):
    '''Main driver for the graphing, each item in the list_of_inputs is a path
    to a json file with a set of runs to plot (and, optionally a name).

    name is specified in one of two ways:
    1.  If the command-line arg has just the json file: "./foo.json", then we
    will derive the name as "foo".
    2.  Using a colon: foobar:./foo.json and the name here is "foobar".
    '''

    fig, ax = plt.subplots()

    for i, input in enumerate(list_of_inputs):
        print(f"Working on {os.path.basename(input)}")

        # check if input has a colon and if so, split on it
        if ":" in input:
            name, json = input.split(":")
        else:
            json = input
            name = os.path.splitext(os.path.basename(input))[0]

        indices, data = read_to_array(config, json)

        mean = data.mean(0)
        std = data.std(0)

        line, =  ax.plot(indices, mean, color=config.color(i))
        ax.fill_between(indices, mean - std, mean + std, color=config.color(i),
                        alpha=config.error_alpha)

        line.set_label(name)

    plt.title(config.title)
    plt.xlabel(config.xlabel)
    plt.ylabel(config.ylabel)
    plt.legend()
    plt.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="plot_run_scalar", add_help=True)

    parser.add_argument("inputs", nargs="+",
                        help="the input json files with one per curve")
    parser.add_argument("--field", "-f", type=str, default=None,
                        help="if data is a dictionary, use this key to get values for plotting")
    parser.add_argument("--xlabel", "-x", type=str, default=None,
                        help="string to use as the label of the x-axis.")
    parser.add_argument("--ylabel", "-y", type=str, default=None,
                        help="string to use as the label of the y-axis.")
    parser.add_argument("--title", "-t", type=str, default=None,
                        help="string to use as the title.")

    args = parser.parse_args()

    config = Config(args.__dict__)
    main_plot(config, args.inputs)



"""
Utilities for experiments.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
import pandas as pd
import time

import numpy as np
import torch


device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
cpu = torch.device("cpu")


class Timer:
    def __init__(self, synch=None):
        self.synch = synch or (lambda: None)
        self.synch()
        self.times = [time.perf_counter()]
        self.total_time = 0.0

    def __call__(self, include_in_total=True):
        self.synch()
        self.times.append(time.perf_counter())
        delta_t = self.times[-1] - self.times[-2]
        if include_in_total:
            self.total_time += delta_t
        return delta_t


sep = "/"


def split(path):
    i = path.rfind(sep) + 1
    return path[:i].rstrip(sep), path[i:]


def normpath(path):
    # simplified os.path.normpath
    parts = []
    for p in path.split(sep):
        if p == "..":
            parts.pop()
        elif p.startswith(sep):
            parts = [p]
        else:
            parts.append(p)
    return sep.join(parts)


union = lambda *dicts: {k: v for d in dicts for (k, v) in d.items()}


def path_iter(nested_dict, pfx=()):
    for name, val in nested_dict.items():
        if isinstance(val, dict):
            yield from path_iter(val, (*pfx, name))
        else:
            yield ((*pfx, name), val)


def map_nested(func, nested_dict):
    return {
        k: map_nested(func, v) if isinstance(v, dict) else func(v)
        for k, v in nested_dict.items()
    }


def group_by_key(items):
    res = defaultdict(list)
    for k, v in items:
        res[k].append(v)
    return res


default_table_formats = {
    float: "{:{w}.4f}",
    str: "{:>{w}s}",
    "default": "{:{w}}",
    "title": "{:>{w}s}",
}


def table_formatter(val, is_title=False, col_width=12, formats=None):
    formats = formats or default_table_formats
    type_ = (
        lambda val: float if isinstance(val, (float, np.float)) else type(val)
    )
    return (
        formats["title"]
        if is_title
        else formats.get(type_(val), formats["default"])
    ).format(val, w=col_width)


def every(n, col):
    return lambda data: data[col] % n == 0


class Table:
    def __init__(
        self, keys=None, report=(lambda data: True), formatter=table_formatter
    ):
        self.keys, self.report, self.formatter = keys, report, formatter
        self.log = []

    def append(self, data):
        self.log.append(data)
        data = {" ".join(p): v for p, v in path_iter(data)}
        self.keys = self.keys or data.keys()
        if len(self.log) is 1:
            print(*(self.formatter(k, True) for k in self.keys))
        if self.report(data):
            print(*(self.formatter(data[k]) for k in self.keys))

    def df(self):
        return pd.DataFrame(
            [{"_".join(p): v for p, v in path_iter(row)} for row in self.log]
        )



"""
This trains a cifar-10 under noisy labels with fixes via fiftyone.
TODO: THIS IS NOT YET MAPPED TO FIFTYONE

After loading the data, a fraction of the annotations will be corrupted at
random.  Then, the training loop will ensue; after every training (full), a
fraction of the samples will be evaluated for erroneous annotations and fixed
(by using the original, correct ones)

What is the comparison?
To allow for a comparison, only half of the initial data will be used; at
each iteration through the main loop, a fraction of this data will be added.
For the annotation error case, we will fix the annotations and add the new
data.  For the baseline, we will just add the new data, which is what a naive
data scientist would do anyway.

TODO Another comparison would be to select just random samples to check and fix.

In addition, simple reporting on the accuracy of the identification of errors
in the annotations

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from functools import partial
import json
import random
import sys
import time

from scipy.stats import entropy

from config import *
from datasets import *
from simple_resnet import *
from utils import Timer


localtime = lambda: time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())


def main(config):

    ## Initial Data Input
    # Productes train_set and valid_set that are lists of tuples: (image, label)
    whole_dataset = cifar10(root=DATA_DIR)
    timer = Timer()
    print("Preprocessing training data")
    transforms = [
        partial(
            normalise,
            mean=np.array(cifar10_mean, dtype=np.float32),
            std=np.array(cifar10_std, dtype=np.float32),
        ),
        partial(transpose, source="NHWC", target="NCHW"),
    ]
    whole_train_set = list(
        zip(
            *preprocess(
                whole_dataset["train"], [partial(pad, border=4)] + transforms
            ).values()
        )
    )
    print(f"Finished in {timer():.2} seconds")
    print("Preprocessing test data")
    valid_set = list(
        zip(*preprocess(whole_dataset["valid"], transforms).values())
    )
    print(f"Finished in {timer():.2} seconds")

    # function of dataset
    N_labels = 10

    # set up the variables for training the model in each increment of the dataset size
    lr_schedule = PiecewiseLinear([0, 5, config.epochs], [0, 0.4, 0])
    train_transforms = [Crop(32, 32), FlipLR(), Cutout(8, 8)]

    # compute the derived parameters for the trial based on the dataset and the
    # provided config.
    total_N = len(whole_train_set)

    # should be cleaner
    if config.n_max < 0:
        config.n_max = total_N

    start_N = round(config.p_initial * total_N)

    incr_N = round((config.n_max - start_N) / config.n_increases)

    corrupt_N = round(config.p_corrupt * total_N)

    print(f"Setting up the experiment: {total_N} training samples.")
    print(f"- starting with {start_N}")
    print(f"- incrementing by {incr_N} for {config.n_increases} rounds")

    total_stats = {}

    for run_index in range(config.runs):

        print(f"Starting training iteration {run_index} at {localtime()}")

        # first, corrupt a random percentage of the data
        # this is wasteful and will create a copy of the training data in memory
        corr_train_set = whole_train_set.copy()

        # the strategy for corrupting the data randomly and then recalling the
        # original values
        corr_bool = [True for x in range(corrupt_N)]
        corr_bool.extend([False for x in range(total_N - corrupt_N)])

        for i in range(corrupt_N):
            t = corr_train_set[i]
            c = random.randint(0, N_labels - 1)
            while c == t[1]:
                c = random.randint(0, N_labels - 1)
            corr_train_set[i] = (t[0], c)

        shuffler = list(range(total_N))
        random.shuffle(shuffler)

        # common randomization of lists by way of an index permutation (shuffler)
        # original: a
        # shuffled: s
        # shuffler on indices: shuffler
        # s = [a[x] for x in shuffler]
        # supports --> s[i] = a[shuffler[i]]
        corr_train_set = [corr_train_set[x] for x in shuffler]
        corr_bool = [corr_bool[x] for x in shuffler]

        inuse_N = start_N

        model = Network(simple_resnet()).to(device).half()
        logs, state = Table(), {MODEL: model, LOSS: x_ent_loss}

        valid_batches = DataLoader(
            valid_set, config.batch_size, shuffle=False, drop_last=False
        )

        # initially randomly shuffle the dataset and take the initial number of samples
        corr_train_set_use = corr_train_set[0:inuse_N]
        corr_train_set_avail = corr_train_set[inuse_N:]
        print(
            f"Split training set into two; using {len(corr_train_set_use)}, available {len(corr_train_set_avail)}"
        )

        sm = torch.nn.Softmax(dim=1)

        stats = {}

        for iteration in range(config.n_increases):
            print(f"beginning next round of training, using {inuse_N} samples")

            if config.cold_start:
                model = Network(simple_resnet()).to(device).half()
                logs, state = Table(), {MODEL: model, LOSS: x_ent_loss}

            train_batches = DataLoader(
                Transform(corr_train_set_use, train_transforms),
                config.batch_size,
                shuffle=True,
                set_random_choices=True,
                drop_last=True,
            )
            lr = (
                lambda step: lr_schedule(step / len(train_batches))
                / config.batch_size
            )
            opts = [
                SGD(
                    trainable_params(model).values(),
                    {
                        "lr": lr,
                        "weight_decay": Const(5e-4 * config.batch_size),
                        "momentum": Const(0.9),
                    },
                )
            ]
            state[OPTS] = opts

            for epoch in range(config.epochs):
                logs.append(
                    union(
                        {"epoch": epoch + 1},
                        train_epoch(
                            state,
                            Timer(torch.cuda.synchronize),
                            train_batches,
                            valid_batches,
                        ),
                    )
                )
            logs.df().query(f"epoch=={config.epochs}")[
                ["train_acc", "valid_acc"]
            ].describe()

            model.train(False)  # == model.eval()

            # record scores for this iteration
            iteration_stats = {}
            iteration_stats["in_use"] = inuse_N

            correct = 0
            total = 0
            class_correct = list(0.0 for i in range(10))
            class_total = list(0.0 for i in range(10))
            with torch.no_grad():
                for data in valid_batches.dataloader:
                    images, labels = data
                    inputs = dict(input=images.cuda().half())
                    outputs = model(inputs)
                    y = outputs["logits"]
                    _, predicted = torch.max(y, 1)
                    total += labels.size(0)
                    labels_gpu = labels.cuda().half()
                    correct += (predicted == labels_gpu).sum().item()
                    c = (predicted == labels_gpu).squeeze()
                    for i in range(min(config.batch_size, len(labels))):
                        label = labels[i]
                        class_correct[label] += c[i].item()
                        class_total[label] += 1

            iteration_stats["validation_accuracy"] = correct / total

            # how many can we fix?
            fix_N = round(inuse_N * config.p_fixable)

            if config.perfect_fixes:
                # implements an upper-bound-like stat, that only selects incorrect samples
                # assumes `config.make_fixes = True`
                fixed = 0

                for fix_i, needs_fixing in enumerate(corr_bool[:inuse_N]):
                    if not needs_fixing:
                        continue
                    t = corr_train_set_use[fix_i]
                    w = whole_train_set[shuffler[fix_i]][1]
                    corr_train_set_use[fix_i] = (t[0], w)
                    corr_bool[fix_i] = False
                    fixed += 1
                    if fixed >= fix_N:
                        break

                iteration_stats["fix_correct"] = fixed
                iteration_stats["fix_percent_of_N"] = fixed / fix_N
                current_corrupted = sum(corr_bool[:inuse_N])
                iteration_stats["current_corrupted"] = current_corrupted

                print(
                    f'"predicted" {fixed} of {fix_N} proposals to fix correctly'
                )
                print(f"{fixed / fix_N}")
                print(
                    f"there are currently a total of {current_corrupted} corrupted labels in use"
                )
            else:
                review_batches = DataLoader(
                    corr_train_set_use, 1, shuffle=False, drop_last=False
                )
                scores = []
                with torch.no_grad():
                    for data in review_batches.dataloader:
                        images, labels = data
                        inputs = dict(input=images.cuda().half())
                        outputs = model(inputs)
                        y = outputs["logits"]
                        # _, predicted = torch.max(y, 1)

                        # likelihood of error in annotation can be a number of
                        # different heuristics.  just use entropy for now to get the
                        # whole script going

                        # todo: implement an entropy in torch to work on gpu
                        z = sm(y).cpu().numpy().squeeze()
                        # e  = entropy(sm(y).numpy())
                        e = entropy(z)
                        scores.append(e)

                scores = np.asarray(scores)
                si = np.argsort(scores)
                si = np.flip(si.squeeze())  # because higher entropy at front
                # beware of a need to sort in other
                # heuristics

                # how many of those that we want to fix need to be fixed
                # this is a bit of a funky calculation because we may want to fix more
                # than we need to if we've already fixed some --> so, just count the
                # true positives for now and neglect any false positives
                fix_correct = 0
                for fix_i in si[:fix_N]:
                    if corr_bool[fix_i]:
                        fix_correct += 1

                iteration_stats["fix_correct"] = fix_correct
                iteration_stats["fix_percent_of_N"] = fix_correct / fix_N

                current_corrupted = sum(corr_bool[:inuse_N])
                iteration_stats["current_corrupted"] = current_corrupted

                print(
                    f"predicted {fix_correct} of {fix_N} proposals to fix correctly"
                )
                print(f"{fix_correct / fix_N}")
                print(
                    f"there are currently a total of {current_corrupted} corrupted labels in use"
                )

                if config.make_fixes:
                    print(f"making the fixes")
                    # fix the corruptions
                    for fix_i in si[:fix_N]:
                        t = corr_train_set_use[fix_i]
                        w = whole_train_set[shuffler[fix_i]][1]
                        corr_train_set_use[fix_i] = (t[0], w)
                        corr_bool[fix_i] = False
                else:
                    print(f"not making the fixes because of the config")

            model.train(True)

            # extend the corr_train_set_use with that from avail
            corr_train_set_use.extend(corr_train_set_avail[0:incr_N])
            corr_train_set_avail = corr_train_set_avail[incr_N:]
            inuse_N += incr_N
            assert inuse_N == len(corr_train_set_use)

            stats[inuse_N] = iteration_stats

        total_stats[run_index] = stats

    print(f"finished the full training; stats to follow")
    print(total_stats)

    if config.stats_path:
        with open(config.stats_path, "w") as fp:
            json.dump(total_stats, fp)


if __name__ == "__main__":

    config = commandline()

    print(f"running with config: {config}")

    main(config)



"""
Script to generate a screenshot of some object detections from the BDD dataset.

You need to run this inside of an interactive shell like IPython because the
system will launch the dashboard and we need to have that process independent
of the python code.

Outputs are placed in the `outputs/` directory.

Usage::

    # From inside IPython
    run shot_bdd_detections.py

Copyright 2017-2023, Voxel51, Inc.
voxel51.com
"""
import os
import time

import eta.core.utils as etau
import eta.core.storage as etas

import fiftyone as fo

import fiftyone.brain as fob


FILE_ID = "1uioZHQ8VVUYb2OIrdD5M5gSw4XmGIOi_"

DATASET_DIR = "data/bdd_subset"
OUTPUT_DIR = "outputs"


print("Working on BDD object detections")

# Download dataset if necessary
if not os.path.isdir(DATASET_DIR):
    print("Downloading dataset to '%s'" % DATASET_DIR)
    client = etas.GoogleDriveStorageClient()
    tmp_path = client.get_file_metadata(FILE_ID)["name"]
    client.download(FILE_ID, tmp_path)
    etau.extract_archive(tmp_path, outdir="data", delete_archive=True)

dataset = fo.Dataset.from_image_detection_dataset(DATASET_DIR, "BDD")

fob.compute_uniqueness(dataset)

view = dataset.view().sort_by("uniqueness", reverse=True)

print("Launching dashboard...")
session = fo.launch_dashboard(view=view)

input("Press enter when you are ready for the screenshot:")

localtime = lambda: time.strftime("%Y%m%d-%H%M%S", time.localtime())
outname = "bdd_detections_" + localtime() + ".png"
outpath = os.path.join(OUTPUT_DIR, outname)

etau.save_window_snapshot("FiftyOne", outpath)



#!/usr/bin/env python
"""
Installs `fiftyone-brain`.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import os
from setuptools import setup


with open("PYPI_README.md", "r") as fh:
    long_description = fh.read()

with open("LICENSE", "r") as fh:
    long_description += "\n## License\n\n" + fh.read()


VERSION = "0.11.0"


def get_version():
    if "RELEASE_VERSION" in os.environ:
        version = os.environ["RELEASE_VERSION"]
        if not version.startswith(VERSION):
            raise ValueError(
                "Release version doest not match version: %s and %s"
                % (version, VERSION)
            )
        return version

    return VERSION


setup(
    name="fiftyone-brain",
    version=get_version(),
    description="FiftyOne Brain",
    author="Voxel51, Inc.",
    author_email="info@voxel51.com",
    url="https://github.com/voxel51/fiftyone-brain",
    license="Freeware (Custom)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["fiftyone.brain"],
    include_package_data=True,
    install_requires=["numpy", "scipy>=1.2.0", "scikit-learn"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: Freeware",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    scripts=[],
    python_requires=">=3.7",
)



"""
Piencone similarity backend.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np

import eta.core.utils as etau

import fiftyone.core.utils as fou
from fiftyone.brain.similarity import (
    SimilarityConfig,
    Similarity,
    SimilarityIndex,
)
import fiftyone.brain.internal.core.utils as fbu

pinecone = fou.lazy_import("pinecone")


logger = logging.getLogger(__name__)

_SUPPORTED_METRICS = ("cosine", "dotproduct", "euclidean")


class PineconeSimilarityConfig(SimilarityConfig):
    """Configuration for the Pinecone similarity backend.

    Args:
        embeddings_field (None): the sample field containing the embeddings,
            if one was provided
        model (None): the :class:`fiftyone.core.models.Model` or name of the
            zoo model that was used to compute embeddings, if known
        patches_field (None): the sample field defining the patches being
            analyzed, if any
        supports_prompts (None): whether this run supports prompt queries
        index_name (None): the name of a Pinecone index to use or create. If
            none is provided, a new index will be created
        index_type (None): the index type to use when creating a new index
        namespace (None): a namespace under which to store vectors added to the
            index
        metric (None): the embedding distance metric to use when creating a
            new index. Supported values are
            ``("cosine", "dotproduct", "euclidean")``
        replicas (None): an optional number of replicas when creating a new
            index
        shards (None): an optional number of shards when creating a new index
        pods (None): an optional number of pods when creating a new index
        pod_type (None): an optional pod type when creating a new index
        api_key (None): a Pinecone API key to use
        environment (None): a Pinecone environment to use
        project_name (None): a Pinecone project to use
    """

    def __init__(
        self,
        embeddings_field=None,
        model=None,
        patches_field=None,
        supports_prompts=None,
        index_name=None,
        index_type=None,
        namespace=None,
        metric=None,
        replicas=None,
        shards=None,
        pods=None,
        pod_type=None,
        api_key=None,
        environment=None,
        project_name=None,
        **kwargs,
    ):
        if metric is not None and metric not in _SUPPORTED_METRICS:
            raise ValueError(
                "Unsupported metric '%s'. Supported values are %s"
                % (metric, _SUPPORTED_METRICS)
            )

        super().__init__(
            embeddings_field=embeddings_field,
            model=model,
            patches_field=patches_field,
            supports_prompts=supports_prompts,
            **kwargs,
        )

        self.index_name = index_name
        self.index_type = index_type
        self.namespace = namespace
        self.metric = metric
        self.replicas = replicas
        self.shards = shards
        self.pods = pods
        self.pod_type = pod_type

        # store privately so these aren't serialized
        self._api_key = api_key
        self._environment = environment
        self._project_name = project_name

    @property
    def method(self):
        return "pinecone"

    @property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, value):
        self._api_key = value

    @property
    def environment(self):
        return self._environment

    @environment.setter
    def environment(self, value):
        self._environment = value

    @property
    def project_name(self):
        return self._project_name

    @project_name.setter
    def project_name(self, value):
        self._project_name = value

    @property
    def max_k(self):
        return 10000  # Pinecone limit

    @property
    def supports_least_similarity(self):
        return False

    @property
    def supported_aggregations(self):
        return ("mean",)

    def load_credentials(
        self, api_key=None, environment=None, project_name=None
    ):
        self._load_parameters(
            api_key=api_key, environment=environment, project_name=project_name
        )


class PineconeSimilarity(Similarity):
    """Pinecone similarity factory.

    Args:
        config: a :class:`PineconeSimilarityConfig`
    """

    def ensure_requirements(self):
        fou.ensure_package("pinecone-client")

    def ensure_usage_requirements(self):
        fou.ensure_package("pinecone-client")

    def initialize(self, samples, brain_key):
        return PineconeSimilarityIndex(
            samples, self.config, brain_key, backend=self
        )


class PineconeSimilarityIndex(SimilarityIndex):
    """Class for interacting with Pinecone similarity indexes.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`PineconeSimilarityConfig` used
        brain_key: the brain key
        backend (None): a :class:`PineconeSimilarity` instance
    """

    def __init__(self, samples, config, brain_key, backend=None):
        super().__init__(samples, config, brain_key, backend=backend)
        self._index = None
        self._initialize()

    def _initialize(self):
        pinecone.init(
            api_key=self.config.api_key,
            environment=self.config.environment,
            project_name=self.config.project_name,
        )

        try:
            index_names = pinecone.list_indexes()
        except Exception as e:
            raise ValueError(
                "Failed to connect to Pinecone backend at environment '%s'. "
                "Refer to https://docs.voxel51.com/integrations/pinecone.html "
                "for more information" % self.config.environment
            ) from e

        if self.config.index_name is None:
            root = "fiftyone-" + fou.to_slug(self.samples._root_dataset.name)
            index_name = fbu.get_unique_name(root, index_names)

            self.config.index_name = index_name
            self.save_config()

        if self.config.index_name in index_names:
            index = pinecone.Index(self.config.index_name)
        else:
            index = None

        self._index = index

    def _create_index(self, dimension):
        kwargs = dict(
            index_type=self.config.index_type,
            metric=self.config.metric,
            replicas=self.config.replicas,
            shards=self.config.shards,
            pods=self.config.pods,
            pod_type=self.config.pod_type,
        )
        kwargs = {k: v for k, v in kwargs.items() if v is not None}

        pinecone.create_index(
            self.config.index_name,
            dimension=dimension,
            **kwargs,
        )

        self._index = pinecone.Index(self.config.index_name)

    @property
    def index(self):
        """The ``pinecone.Index`` instance for this index."""
        return self._index

    @property
    def total_index_size(self):
        if self._index is None:
            return None

        return self._index.describe_index_stats()["total_vector_count"]

    def add_to_index(
        self,
        embeddings,
        sample_ids,
        label_ids=None,
        overwrite=True,
        allow_existing=True,
        warn_existing=False,
        batch_size=100,
        namespace=None,
    ):
        if namespace is None:
            namespace = self.config.namespace

        if self._index is None:
            self._create_index(embeddings.shape[1])

        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if warn_existing or not allow_existing or not overwrite:
            existing_ids = self._get_existing_ids(ids)
            num_existing = len(existing_ids)

            if num_existing > 0:
                if not allow_existing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that already exist in the index"
                        % (num_existing, next(iter(existing_ids)))
                    )

                if warn_existing:
                    if overwrite:
                        logger.warning(
                            "Overwriting %d IDs that already exist in the "
                            "index",
                            num_existing,
                        )
                    else:
                        logger.warning(
                            "Skipping %d IDs that already exist in the index",
                            num_existing,
                        )
        else:
            existing_ids = set()

        if existing_ids and not overwrite:
            del_inds = [i for i, _id in enumerate(ids) if _id in existing_ids]

            embeddings = np.delete(embeddings, del_inds)
            sample_ids = np.delete(sample_ids, del_inds)
            if label_ids is not None:
                label_ids = np.delete(label_ids, del_inds)

        embeddings = [e.tolist() for e in embeddings]
        sample_ids = list(sample_ids)
        if label_ids is not None:
            ids = list(label_ids)
        else:
            ids = list(sample_ids)

        for _embeddings, _ids, _sample_ids in zip(
            fou.iter_batches(embeddings, batch_size),
            fou.iter_batches(ids, batch_size),
            fou.iter_batches(sample_ids, batch_size),
        ):
            _id_dicts = [
                {"id": _id, "sample_id": _sid}
                for _id, _sid in zip(_ids, _sample_ids)
            ]
            self._index.upsert(
                list(zip(_ids, _embeddings, _id_dicts)),
                namespace=namespace,
            )

    def remove_from_index(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if not allow_missing or warn_missing:
            existing_ids = self._index.fetch(ids).vectors.keys()
            missing_ids = set(existing_ids) - set(ids)
            num_missing = len(missing_ids)

            if num_missing > 0:
                if not allow_missing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that are not present in the "
                        "index" % (num_missing, missing_ids[0])
                    )

                if warn_missing:
                    logger.warning(
                        "Ignoring %d IDs that are not present in the index",
                        num_missing,
                    )

        self._index.delete(ids=ids)

    def get_embeddings(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        if label_ids is not None:
            if self.config.patches_field is None:
                raise ValueError("This index does not support label IDs")

            if sample_ids is not None:
                logger.warning(
                    "Ignoring sample IDs when label IDs are provided"
                )

        if sample_ids is not None and self.config.patches_field is not None:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_patch_embeddings_from_sample_ids(sample_ids)
        elif self.config.patches_field is not None:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_patch_embeddings_from_label_ids(label_ids)
        else:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_sample_embeddings(sample_ids)

        num_missing_ids = len(missing_ids)
        if num_missing_ids > 0:
            if not allow_missing:
                raise ValueError(
                    "Found %d IDs (eg %s) that do not exist in the index"
                    % (num_missing_ids, missing_ids[0])
                )

            if warn_missing:
                logger.warning(
                    "Skipping %d IDs that do not exist in the index",
                    num_missing_ids,
                )

        embeddings = np.array(embeddings)
        sample_ids = np.array(sample_ids)
        if label_ids is not None:
            label_ids = np.array(label_ids)

        return embeddings, sample_ids, label_ids

    def cleanup(self):
        pinecone.delete_index(self.config.index_name)
        self._index = None

    def _get_existing_ids(self, ids, batch_size=1000):
        existing_ids = set()
        for batch_ids in fou.iter_batches(ids, batch_size):
            response = self._index.fetch(ids=list(batch_ids))["vectors"]
            existing_ids.update(response.keys())

        return existing_ids

    def _get_sample_embeddings(self, sample_ids, batch_size=1000):
        found_embeddings = []
        found_sample_ids = []

        if sample_ids is None:
            raise ValueError(
                "Pinecone does not support retrieving all vectors in an index"
            )

        for batch_ids in fou.iter_batches(sample_ids, batch_size):
            response = self._index.fetch(ids=list(batch_ids))["vectors"]

            for r in response.values():
                found_embeddings.append(r["values"])
                found_sample_ids.append(r["id"])

        missing_ids = list(set(sample_ids) - set(found_sample_ids))

        return found_embeddings, found_sample_ids, None, missing_ids

    def _get_patch_embeddings_from_label_ids(self, label_ids, batch_size=1000):
        found_embeddings = []
        found_sample_ids = []
        found_label_ids = []

        if label_ids is None:
            raise ValueError(
                "Pinecone does not support retrieving all vectors in an index"
            )

        for batch_ids in fou.iter_batches(label_ids, batch_size):
            response = self._index.fetch(ids=list(batch_ids))["vectors"]

            for r in response.values():
                found_embeddings.append(r["values"])
                found_sample_ids.append(r["metadata"]["sample_id"])
                found_label_ids.append(r["id"])

        missing_ids = list(set(label_ids) - set(found_label_ids))

        return found_embeddings, found_sample_ids, found_label_ids, missing_ids

    def _get_patch_embeddings_from_sample_ids(
        self, sample_ids, batch_size=100
    ):
        found_embeddings = []
        found_sample_ids = []
        found_label_ids = []

        query_vector = [0.0] * self._index.describe_index_stats().dimension
        top_k = min(batch_size, self.config.max_k)

        for batch_ids in fou.iter_batches(sample_ids, batch_size):
            response = self._index.query(
                vector=query_vector,
                filter={"sample_id": {"$in": list(batch_ids)}},
                top_k=top_k,
                include_values=True,
                include_metadata=True,
            )

            for r in response["matches"]:
                found_embeddings.append(r["values"])
                found_sample_ids.append(r["metadata"]["sample_id"])
                found_label_ids.append(r["id"])

        missing_ids = list(set(sample_ids) - set(found_sample_ids))

        return found_embeddings, found_sample_ids, found_label_ids, missing_ids

    def _kneighbors(
        self,
        query=None,
        k=None,
        reverse=False,
        aggregation=None,
        return_dists=False,
    ):
        if query is None:
            raise ValueError("Pinecone does not support full index neighbors")

        if reverse is True:
            raise ValueError(
                "Pinecone does not support least similarity queries"
            )

        if k is None or k > self.config.max_k:
            raise ValueError("Pincone requires k<=%s" % self.config.max_k)

        if aggregation not in (None, "mean"):
            raise ValueError("Unsupported aggregation '%s'" % aggregation)

        query = self._parse_neighbors_query(query)
        if aggregation == "mean" and query.ndim == 2:
            query = query.mean(axis=0)

        single_query = query.ndim == 1
        if single_query:
            query = [query]

        if self.config.patches_field is not None:
            index_ids = self.current_label_ids
        else:
            index_ids = self.current_sample_ids

        _filter = {"id": {"$in": list(index_ids)}}

        ids = []
        dists = []
        for q in query:
            response = self._index.query(
                vector=q.tolist(), top_k=k, filter=_filter
            )
            ids.append([r["id"] for r in response["matches"]])
            if return_dists:
                dists.append([r["score"] for r in response["matches"]])

        if single_query:
            ids = ids[0]
            if return_dists:
                dists = dists[0]

        if return_dists:
            return ids, dists

        return ids

    def _parse_neighbors_query(self, query):
        if etau.is_str(query):
            query_ids = [query]
            single_query = True
        else:
            query = np.asarray(query)

            # Query by vector(s)
            if np.issubdtype(query.dtype, np.number):
                return query

            query_ids = list(query)
            single_query = False

        # Query by ID(s)
        response = self._index.fetch(query_ids)["vectors"]
        query = np.array([response[_id]["values"] for _id in query_ids])

        if single_query:
            query = query[0, :]

        return query

    @classmethod
    def _from_dict(cls, d, samples, config, brain_key):
        return cls(samples, config, brain_key)
